from enum import Enum


class Wheel(Enum):
    """
    The name of each wheel on the Magnebot.
    """

    wheel_left_front = 0
    wheel_left_back = 1
    wheel_right_front = 2
    wheel_right_back = 3
